from functools import wraps
from typing import Tuple, List, Optional, Callable

import torch
from torch import nn

from .util import LayerBounds, add_method, AlphaBetas, AlphaBeta, TensorFunction, LayerBound


def alpha_beta(class_or_obj):
    types = {
        nn.Sequential: alpha_beta_sequential,
        nn.Linear: alpha_beta_linear,
        nn.ReLU: alpha_beta_relu,
        nn.Sigmoid: alpha_beta_sigmoid,
        nn.Tanh: alpha_beta_tanh
    }

    for layer_type, alpha_beta_fun in types.items():
        if isinstance(class_or_obj, layer_type) or \
                (isinstance(class_or_obj, type) and issubclass(class_or_obj, layer_type)):
            return alpha_beta_fun(class_or_obj)

    raise NotImplementedError('Selected type of layer not supported')


def add_alpha_beta_submodules(model: nn.Sequential):
    for module in model:
        if not hasattr(module, 'alpha_beta'):
            # Decorator also adds the method inplace.
            alpha_beta(module)


def alpha_beta_sequential(class_or_obj):
    def _alpha_beta(self: nn.Sequential, layer_bounds: LayerBounds, **kwargs) -> AlphaBetas:
        add_alpha_beta_submodules(self)

        alpha_betas = []

        for module, layer_bound in zip(self, layer_bounds):
            ab = module.alpha_beta(layer_bound, **kwargs)
            alpha_betas.append(ab)

        return alpha_betas

    add_method(class_or_obj, 'alpha_beta', _alpha_beta)
    return class_or_obj


def regimes(LB: torch.Tensor, UB: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    n = UB <= 0
    p = 0 <= LB
    np = (LB < 0) & (0 < UB)

    return n, p, np


def assert_bound_order(func):
    @wraps(func)
    def wrapper(self, layer_bound: LayerBound, **kwargs):
        LB, UB = layer_bound
        assert torch.all(LB <= UB + 1e-6)

        return func(self, layer_bound, **kwargs)

    return wrapper


def alpha_beta_relu(class_or_obj):
    @assert_bound_order
    def alpha_beta(self: nn.ReLU, layer_bound: LayerBound, adaptive_relu: bool = True, **kwargs) -> AlphaBeta:
        """
        See the source below for information on what adaptive ReLU means and how regimes of the input
        are exploited to compute bounds.

        @misc{zhang2018efficient,
              title={Efficient Neural Network Robustness Certification with General Activation Functions},
              author={Huan Zhang and Tsui-Wei Weng and Pin-Yu Chen and Cho-Jui Hsieh and Luca Daniel},
              year={2018},
              eprint={1811.00866},
              archivePrefix={arXiv},
              primaryClass={cs.LG}
        }

        :param self:
        :param layer_bound:
        :param adaptive_relu:
        :param kwargs:
        :return:
        """
        LB, UB = layer_bound
        n, p, np = regimes(LB, UB)

        alpha_lower_k = torch.zeros_like(LB)
        alpha_upper_k = torch.zeros_like(LB)
        beta_lower_k = torch.zeros_like(LB)
        beta_upper_k = torch.zeros_like(LB)

        alpha_lower_k[n] = 0
        alpha_upper_k[n] = 0
        beta_lower_k[n] = 0
        beta_upper_k[n] = 0

        alpha_lower_k[p] = 1
        alpha_upper_k[p] = 1
        beta_lower_k[p] = 0
        beta_upper_k[p] = 0

        LB, UB = LB[np], UB[np]

        z = UB / (UB - LB)
        if adaptive_relu:
            # Utilize that bool->float conversion is true=1 and false=0
            a = (UB >= torch.abs(LB)).to(torch.float)
        else:
            a = z

        alpha_lower_k[np] = a
        alpha_upper_k[np] = z
        beta_lower_k[np] = 0
        beta_upper_k[np] = -LB * z

        return (alpha_lower_k, alpha_upper_k), (beta_lower_k, beta_upper_k)

    add_method(class_or_obj, 'alpha_beta', alpha_beta)
    return class_or_obj


def alpha_beta_sigmoid(class_or_obj):
    def derivative(d):
        return torch.sigmoid(d) * (1 - torch.sigmoid(d))

    @assert_bound_order
    def alpha_beta(self: nn.Sigmoid, layer_bound: LayerBound, **kwargs) -> AlphaBeta:
        return alpha_beta_general(layer_bound, torch.sigmoid, derivative)

    add_method(class_or_obj, 'alpha_beta', alpha_beta)
    return class_or_obj


def alpha_beta_tanh(class_or_obj):
    def derivative(d):
        return 1 - torch.tanh(d) ** 2

    @assert_bound_order
    def alpha_beta(self: nn.Tanh, layer_bound: LayerBound, **kwargs) -> AlphaBeta:
        return alpha_beta_general(layer_bound, torch.tanh, derivative)

    add_method(class_or_obj, 'alpha_beta', alpha_beta)
    return class_or_obj


def alpha_beta_general(layer_bound: LayerBound, func: TensorFunction, derivative: TensorFunction) -> AlphaBeta:
    """
    Function to compute upper and lower bounds for S-shaped activation functions.
    Assumptions:
    - :LB < :UB
    - :derivative is the derivative of :func
    - :func is S-/sigmoid shaped

    See the source below for information on how regimes of the input are exploited to compute bounds.

    @misc{zhang2018efficient,
          title={Efficient Neural Network Robustness Certification with General Activation Functions},
          author={Huan Zhang and Tsui-Wei Weng and Pin-Yu Chen and Cho-Jui Hsieh and Luca Daniel},
          year={2018},
          eprint={1811.00866},
          archivePrefix={arXiv},
          primaryClass={cs.LG}
    }

    :param layer_bound:
    :param func:
    :param derivative:
    :return:
    """
    LB, UB = layer_bound
    n, p, np = regimes(LB, UB)

    alpha_lower = torch.zeros_like(LB)
    alpha_upper = torch.zeros_like(LB)
    beta_lower = torch.zeros_like(LB)
    beta_upper = torch.zeros_like(LB)

    LB_act, UB_act = func(LB), func(UB)
    LB_prime, UB_prime = derivative(LB), derivative(UB)

    d = (LB + UB) * 0.5  # Let d be the midpoint of the two bounds
    d_act = func(d)
    d_prime = derivative(d)

    slope = (UB_act - LB_act) / (UB - LB)

    def add_linear(alpha, beta, mask, a, x, y, a_mask=True):
        if a_mask:
            a = a[mask]

        alpha[mask] = a
        beta[mask] = y[mask] - a * x[mask]

    ###################
    # Negative regime #
    ###################
    # Upper bound
    # - Exact slope between LB and UB
    add_linear(alpha_upper, beta_upper, mask=n, a=slope, x=UB, y=UB_act)

    # Lower bound
    # - d = (LB + UB) / 2 for midpoint
    # - Slope is sigma'(d) and it has to cross through sigma(d)
    add_linear(alpha_lower, beta_lower, mask=n, a=d_prime, x=d, y=d_act)

    ###################
    # Positive regime #
    ###################
    # Lower bound
    # - Exact slope between LB and UB
    add_linear(alpha_lower, beta_lower, mask=p, a=slope, x=LB, y=LB_act)

    # Upper bound
    # - d = (LB + UB) / 2 for midpoint
    # - Slope is sigma'(d) and it has to cross through sigma(d)
    add_linear(alpha_upper, beta_upper, mask=p, a=d_prime, x=d, y=d_act)

    #################
    # Crossing zero #
    #################
    # Upper bound #
    # If tangent to UB is below LB, then take direct slope between LB and UB
    direct_upper = np & (slope <= UB_prime)
    add_linear(alpha_upper, beta_upper, mask=direct_upper, a=slope, x=LB, y=LB_act)

    # Else use bisection to find upper bound on slope.
    implicit_upper = np & (slope > UB_prime)

    def f_upper(d: torch.Tensor) -> torch.Tensor:
        a_slope = (func(d) - func(LB[implicit_upper])) / (d - LB[implicit_upper])
        a_derivative = derivative(d)
        return a_slope - a_derivative

    # Bisection will return left and right bounds for d s.t. f_upper(d) is zero
    # Derivative of left bound will overapproximate the slope - hence a true bound
    d_upper, _ = bisection(torch.zeros_like(UB[implicit_upper]), UB[implicit_upper], f_upper)
    # Slope has to attach to (LB, sigma(LB))
    add_linear(alpha_upper, beta_upper, mask=implicit_upper, a=derivative(d_upper), x=LB, y=LB_act, a_mask=False)

    # Lower bound #
    # If tangent to LB is above UB, then take direct slope between LB and UB
    direct_lower = np & (slope <= LB_prime)
    add_linear(alpha_lower, beta_lower, mask=direct_lower, a=slope, x=UB, y=UB_act)

    # Else use bisection to find upper bound on slope.
    implicit_lower = np & (slope > LB_prime)

    def f_lower(d: torch.Tensor) -> torch.Tensor:
        a_slope = (func(UB[implicit_lower]) - func(d)) / (UB[implicit_lower] - d)
        a_derivative = derivative(d)
        return a_derivative - a_slope

    # Bisection will return left and right bounds for d s.t. f_lower(d) is zero
    # Derivative of right bound will overapproximate the slope - hence a true bound
    _, d_lower = bisection(LB[implicit_lower], torch.zeros_like(LB[implicit_lower]), f_lower)
    # Slope has to attach to (UB, sigma(UB))
    add_linear(alpha_lower, beta_lower, mask=implicit_lower, a=derivative(d_lower), x=UB, y=UB_act, a_mask=False)

    return (alpha_lower, alpha_upper), (beta_lower, beta_upper)


def bisection(l: torch.Tensor, h: torch.Tensor, f: TensorFunction, num_iter: int = 20) -> Tuple[torch.Tensor, torch.Tensor]:
    midpoint = (l + h) / 2

    for _ in range(num_iter):
        y = f(midpoint)

        l[y <= 0] = midpoint[y <= 0]
        h[y > 0] = midpoint[y > 0]

        midpoint = (l + h) / 2

    return l, h


def alpha_beta_linear(class_or_obj):
    def alpha_beta(self: nn.Linear, layer_bound: LayerBound, **kwargs) -> AlphaBeta:
        return (None, None), (None, None)

    add_method(class_or_obj, 'alpha_beta', alpha_beta)
    return class_or_obj



import torch
from torch import nn

from .ibp import ibp
from .crown import linear_bounds, interval_bounds, output_size, crown
from .util import add_method, LinearBounds, IntervalBounds


def crown_ibp(model):
    def crown_ibp_linear(self: nn.Sequential, lower: torch.Tensor, upper: torch.Tensor, **kwargs) -> LinearBounds:
        ibp(self)
        crown(self)

        layer_bounds = self.ibp(lower, upper, pre=True)
        alpha_betas = self.alpha_beta(layer_bounds)

        batch_size = lower.size(0)
        out_size = output_size(lower.size(-1), self)
        bounds = linear_bounds(self, alpha_betas, batch_size, out_size, lower.device, **kwargs)

        return bounds

    add_method(model, 'crown_ibp_linear', crown_ibp_linear)

    def crown_ibp_interval(self: nn.Sequential, lower: torch.Tensor, upper: torch.Tensor, **kwargs) -> IntervalBounds:
        bounds = crown_ibp_linear(self, lower, upper, **kwargs)
        return interval_bounds(bounds, (lower, upper))

    add_method(model, 'crown_ibp_interval', crown_ibp_interval)

    return model



from typing import Callable, Tuple, List, Union, Optional

import torch
from torch import nn

from .alpha_beta import alpha_beta, add_alpha_beta_submodules
from .util import add_method, LinearBounds, IntervalBounds, LayerBounds, AlphaBetas, LinearBound, AlphaBeta, WeightBias


def crown(model: nn.Sequential):
    def crown_linear(self: nn.Sequential, lower: torch.Tensor, upper: torch.Tensor, **kwargs) -> LinearBounds:
        add_alpha_beta_submodules(self)
        subnetwork_crown(self)

        alpha_betas = []

        for i in range(len(self)):
            # We do this wrt. next layer because if the next layer is linear, we can skip this iteration
            b = self[i].subnetwork_crown(self[:i], alpha_betas, (lower, upper))
            alpha_beta = self[i].alpha_beta(b)
            alpha_betas.append(alpha_beta)

        batch_size = lower.size(0)
        out_size = output_size(lower.size(-1), self)
        bounds = linear_bounds(self, alpha_betas, batch_size, out_size, lower.device, **kwargs)

        return bounds

    add_method(model, 'crown_linear', crown_linear)

    def crown_interval(self: nn.Sequential, lower: torch.Tensor, upper: torch.Tensor, **kwargs) -> IntervalBounds:
        bounds = crown_linear(self, lower, upper, **kwargs)
        return interval_bounds(bounds, (lower, upper))

    add_method(model, 'crown_interval', crown_interval)

    model = alpha_beta(model)

    return model


def subnetwork_crown(model):
    if isinstance(model, nn.Sequential):
        return subnetwork_crown_sequential(model)
    elif isinstance(model, nn.Linear):
        return subnetwork_crown_linear(model)
    else:
        return subnetwork_crown_activation(model)


def subnetwork_crown_sequential(model: nn.Sequential):
    for module in model:
        subnetwork_crown(module)


def subnetwork_crown_linear(class_or_obj):
    def subnetwork_crown(self: nn.Linear, model: nn.Sequential, layer_bounds: LayerBounds, input_bounds: IntervalBounds) -> Optional[IntervalBounds]:
        return None

    add_method(class_or_obj, 'subnetwork_crown', subnetwork_crown)
    return class_or_obj


def subnetwork_crown_activation(class_or_obj):
    def subnetwork_crown(self: nn.Module, model: nn.Sequential, alpha_betas: AlphaBetas, input_bounds: IntervalBounds) -> IntervalBounds:
        device = input_bounds[0].device
        batch_size = input_bounds[0].size(0)
        out_size = output_size(input_bounds[0].size(-1), model)
        bounds = linear_bounds(model, alpha_betas, batch_size, out_size, device)

        return interval_bounds(bounds, input_bounds)

    add_method(class_or_obj, 'subnetwork_crown', subnetwork_crown)
    return class_or_obj


def output_size(input_size: int, model: nn.Sequential) -> int:
    out_size = input_size

    for module in model:
        if isinstance(module, nn.Linear):
            out_size = module.out_features

    return out_size


def interval_bounds(bounds: LinearBounds, input_bounds: IntervalBounds) -> IntervalBounds:
    lower, upper = input_bounds
    lower, upper = lower.unsqueeze(-1), upper.unsqueeze(-1)

    # We can do this instead of finding the Q-norm, as we only deal with perturbation over a hyperrectangular input,
    # and not an arbitrary B_p(epsilon) ball
    # This is essentially:
    # - min_x Omega_0 @ x + Omega_accumulator
    # - max_x Gamma_0 @ x + Gamma_accumulator

    mid = (lower + upper) / 2
    diff = (upper - lower) / 2

    Omega, Gamma = bounds

    if Omega is not None:
        Omega_0, Omega_accumulator = Omega
        min_Omega_x = (Omega_0.matmul(mid) - Omega_0.abs().matmul(diff))[..., 0]
        Omega = min_Omega_x + Omega_accumulator

    if Gamma is not None:
        Gamma_0, Gamma_accumulator = Gamma
        max_Gamma_x = (Gamma_0.matmul(mid) + Gamma_0.abs().matmul(diff))[..., 0]
        Gamma = max_Gamma_x + Gamma_accumulator

    return Omega, Gamma


def linear_bounds(model: nn.Sequential, alpha_betas: AlphaBetas, batch_size: int, out_size: int, device: torch.device,
                  bound_lower=True, bound_upper=True) -> LinearBounds:
    # Compute bounds as two iterations to reduce memory consumption by half
    lower = oneside_linear_bound(model, alpha_betas, batch_size, out_size, device, lower=True) if bound_lower else None
    upper = oneside_linear_bound(model, alpha_betas, batch_size, out_size, device, lower=False) if bound_upper else None

    return lower, upper


def oneside_linear_bound(model: nn.Sequential, alpha_betas: AlphaBetas, batch_size: int, out_size: int, device: torch.device, **kwargs) -> LinearBound:
    W_tilde = torch.eye(out_size, device=device).unsqueeze(0).expand(batch_size, out_size, out_size)
    acc = 0

    # List is necessary around zip to allow reversing
    for module, alpha_beta in reversed(list(zip(model, alpha_betas))):
        if not hasattr(module, 'crown_backward'):
            # Decorator also adds the method inplace.
            crown_backward(module)

        W_tilde, bias = module.crown_backward(W_tilde, alpha_beta, **kwargs)
        acc = acc + bias

    return W_tilde, acc


def crown_backward(class_or_obj):
    types = {
        nn.Linear: crown_backward_linear,
        nn.ReLU: crown_backward_activation,
        nn.Sigmoid: crown_backward_activation,
        nn.Tanh: crown_backward_activation
    }

    for layer_type, crown_backward_fun in types.items():
        if isinstance(class_or_obj, layer_type) or \
                (isinstance(class_or_obj, type) and issubclass(class_or_obj, layer_type)):
            return crown_backward_fun(class_or_obj)

    raise NotImplementedError('Selected type of layer not supported')


def crown_backward_linear(class_or_obj):
    def crown_backward(self: nn.Linear, W_tilde: torch.Tensor, alpha_beta: AlphaBeta, **kwargs) -> WeightBias:
        bias = self.bias
        weight = self.weight

        if bias is None:
            bias_acc = 0
        elif bias.dim() == 1:
            bias_acc = W_tilde.matmul(bias)
        else:
            # This allows stochastic dynamics to be treated as an nn.Linear
            bias = bias.view(bias.size(0), 1, bias.size(-1), 1)
            bias_acc = W_tilde.matmul(bias)[..., 0]

        if weight.dim() == 2:
            W_tilde = W_tilde.matmul(weight)
        else:
            if W_tilde.dim() == 3:
                W_tilde = W_tilde.unsqueeze(0)

            W_tilde = W_tilde.matmul(weight.unsqueeze(1))

        return W_tilde, bias_acc

    add_method(class_or_obj, 'crown_backward', crown_backward)
    return class_or_obj


def crown_backward_activation(class_or_obj):
    def crown_backward(self: nn.Module, W_tilde: torch.Tensor, alpha_beta: AlphaBeta, lower: bool = True, **kwargs) -> WeightBias:
        if lower:
            return act_lower(W_tilde, alpha_beta)
        else:
            return act_upper(W_tilde, alpha_beta)

    add_method(class_or_obj, 'crown_backward', crown_backward)
    return class_or_obj


def act_lower(Omega_tilde: torch.Tensor, alpha_beta: AlphaBeta) -> WeightBias:
    (al_k, au_k), (bl_k, bu_k) = alpha_beta

    bias = torch.sum(Omega_tilde * _theta(Omega_tilde, bl_k, bu_k), dim=-1)
    Omega_tilde = Omega_tilde * _omega(Omega_tilde, al_k, au_k)

    return Omega_tilde, bias


def _theta(Omega_tilde: torch.Tensor, beta_lower: torch.Tensor, beta_upper: torch.Tensor) -> torch.Tensor:
    return torch.where(Omega_tilde < 0, beta_upper.unsqueeze(-2), beta_lower.unsqueeze(-2))


def _omega(Omega_tilde: torch.Tensor, alpha_lower: torch.Tensor, alpha_upper: torch.Tensor) -> torch.Tensor:
    return torch.where(Omega_tilde < 0, alpha_upper.unsqueeze(-2), alpha_lower.unsqueeze(-2))


def act_upper(Gamma_tilde: torch.Tensor, alpha_beta: AlphaBeta) -> WeightBias:
    (al_k, au_k), (bl_k, bu_k) = alpha_beta

    bias = torch.sum(Gamma_tilde * _delta(Gamma_tilde, bl_k, bu_k), dim=-1)
    Gamma_tilde = Gamma_tilde * _lambda(Gamma_tilde, al_k, au_k)

    return Gamma_tilde, bias


def _delta(Gamma_tilde: torch.Tensor, beta_lower: torch.Tensor, beta_upper: torch.Tensor) -> torch.Tensor:
    return torch.where(Gamma_tilde < 0, beta_lower.unsqueeze(-2), beta_upper.unsqueeze(-2))


def _lambda(Gamma_tilde: torch.Tensor, alpha_lower: torch.Tensor, alpha_upper: torch.Tensor) -> torch.Tensor:
    return torch.where(Gamma_tilde < 0, alpha_lower.unsqueeze(-2), alpha_upper.unsqueeze(-2))
